/* -*- c -*- */
/* $Id: idmap.c,v 1.3 2002/03/22 19:47:07 cher Exp $ */

/* Copyright (C) 2001,2002 Alexander Chernov <cher@ispras.ru> */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "idmap.h"

#include <reuse/xalloc.h>

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

static void err(char const *, ...);

enum
{
  LOGIN_FIELD_LEN = 32,
  ID_FIELD_LEN = 32,
  TNAME_FIELD_LEN = 64,
  LOGIN1_FIELD_LEN = 32,
  ROOM1_FIELD_LEN = 32,
  LOGIN2_FIELD_LEN = 32,
  ROOM2_FIELD_LEN = 32
};

struct idmap_elem
{
  char *login;
  int   id;
  char *tname;
  char *login1;
  char *room1;
  char *login2;
  char *room2;
};

struct idmap_elem_arr
{
  int                 a, u;
  struct idmap_elem **v;
};

struct idmap
{
  char *top_comments;
  char *bottom_comments;
  struct idmap_elem_arr items;
  int   id_size;
  int  *id_map;
};

static char b_login[LOGIN_FIELD_LEN];
static char b_id[ID_FIELD_LEN];
static char b_tname[TNAME_FIELD_LEN];
static char b_login1[LOGIN1_FIELD_LEN];
static char b_room1[ROOM1_FIELD_LEN];
static char b_login2[LOGIN2_FIELD_LEN];
static char b_room2[ROOM2_FIELD_LEN];

struct fieldinfo
{
  char *name;
  int   maxlen;
  char *ptr;
};

static struct fieldinfo idmap_fieldinfo [] =
{
  { "login", LOGIN_FIELD_LEN, b_login },
  { "id", ID_FIELD_LEN, b_id },
  { "tname", TNAME_FIELD_LEN, b_tname },
  { "login1", LOGIN1_FIELD_LEN, b_login1 },
  { "room1", ROOM1_FIELD_LEN, b_room1 },
  { "login2", LOGIN2_FIELD_LEN, b_login2 },
  { "room2", ROOM2_FIELD_LEN, b_room2 },

  { 0, 0, 0 }
};

static int
read_fields(char const *s, int separator, const struct fieldinfo *fi,
            char const *file, int line)
{
  int i, j;
  char const *p = s;

  for (i = 0; fi[i].maxlen && *p; i++) {
    j = 0;
    memset(fi[i].ptr, 0, fi[i].maxlen);
    while (*p != 0 && *p != separator && *p != '\n') {
      if (j >= fi[i].maxlen - 1) {
        err("%s: %d: read_fields: field %s is too long", file,
            line, fi[i].name);
        return -1;
      }
      fi[i].ptr[j++] = *p++;
    }
    if (*p == separator || *p == '\n') p++;
  }
  if (fi[i].maxlen != 0) {
    err("%s: %d: read_fields: too few fields read: %d", file, line, i);
    return -1;
  }
  return i;
}

void
free_idmap_elem(struct idmap_elem *pdbe)
{
  if (!pdbe) return;
  xfree(pdbe->login);
  xfree(pdbe->tname);
  xfree(pdbe->login1);
  xfree(pdbe->room1);
  xfree(pdbe->login2);
  xfree(pdbe->room2);
  XMEMZERO(pdbe, 1);
  xfree(pdbe);
}

void
free_idmap(struct idmap *pdb)
{
  int i;

  xfree(pdb->top_comments);
  xfree(pdb->bottom_comments);
  for (i = 0; i < pdb->items.u; i++) {
    free_idmap_elem(pdb->items.v[i]);
  }
  if (pdb->items.a > 0) {
    XMEMZERO(pdb->items.v, pdb->items.a);
  }
  xfree(pdb->items.v);
  xfree(pdb->id_map);
  XMEMZERO(pdb, 1);  
}

struct idmap *
read_idmap(char const *name)
{
  struct idmap *pdb = 0;
  struct idmap_elem *pelem = 0;
  FILE *in = 0;
  int line = 0;
  int len;
  int tid, n, i;
  int state = 0;
  char linebuf[1024];

  if (!(in = fopen(name, "r"))) {
    err("read_idmap: cannot open `%s'", name);
    goto cleanup;
  }
  XCALLOC(pdb, 1);

  while (fgets(linebuf, sizeof(linebuf), in) != NULL) {
    line++;
    len = strlen(linebuf);
    if (len == sizeof(linebuf) - 1 && linebuf[len - 1] != '\n') {
      err("%s: %d: read_idmap: line is too long", name, line);
      goto cleanup;
    }
    if (linebuf[0] == '#') {
      if (state == 0) {
        pdb->top_comments = xstrmerge1(pdb->top_comments, linebuf);
      } else if (state == 2) {
        pdb->bottom_comments = xstrmerge1(pdb->bottom_comments, linebuf);
      }
      if (!strcmp(linebuf, "###\n")) {
        state++;
        if (state == 2)
          pdb->bottom_comments = xstrmerge1(pdb->bottom_comments, linebuf);
      }
      continue;
    }
    if (read_fields(linebuf, ':', idmap_fieldinfo, name, line) < 0)
      goto cleanup;
    if (sscanf(b_id, "%d %n", &tid, &n) != 1 || b_id[n]
        || tid <= 0 || tid > 65535) {
      err("%s: %d: invalid team_id: %s, %d", name, line, b_id, tid);
      goto cleanup;
    }

    XCALLOC(pelem, 1);
    pelem->login = xstrdup(b_login);
    pelem->id = tid;
    pelem->tname = xstrdup(b_tname);
    pelem->login1 = xstrdup(b_login1);
    pelem->room1 = xstrdup(b_room1);
    pelem->login2 = xstrdup(b_login2);
    pelem->room2 = xstrdup(b_room2);
    XEXPAND2(pdb->items);
    pdb->items.v[pdb->items.u++] = pelem;
    pelem = 0;
  }

  fclose(in);

  pdb->id_size = -1;
  for (i = 0; i < pdb->items.u; i++) {
    if (pdb->items.v[i]->id > pdb->id_size)
      pdb->id_size = pdb->items.v[i]->id;
  }
  pdb->id_size++;
  if (pdb->id_size > 0) {
    XCALLOC(pdb->id_map, pdb->id_size);
    for (i = 0; i < pdb->id_size; i++)
      pdb->id_map[i] = -1;
  }
  for (i = 0; i < pdb->items.u; i++) {
    pdb->id_map[pdb->items.v[i]->id] = i;
  }

  return pdb;

 cleanup:
  if (in) fclose(in);
  free_idmap(pdb);
  free_idmap_elem(pelem);
  return 0;
}

int
idmap_id_size(struct idmap *db)
{
  return db->id_size;
}

int
idmap_get_index(struct idmap *db, int teamid)
{
  return db->id_map[teamid];
}

char *
idmap_get_name(struct idmap *db, int teamid)
{
  return db->items.v[db->id_map[teamid]]->tname;
}

char *
idmap_get_login(struct idmap *db, int teamid)
{
  return db->items.v[db->id_map[teamid]]->login;
}

char *
idmap_get_login1(struct idmap *db, int teamid)
{
  return db->items.v[db->id_map[teamid]]->login1;
}

char *
idmap_get_login2(struct idmap *db, int teamid)
{
  return db->items.v[db->id_map[teamid]]->login2;
}

#if 0
int
main(int argc, char *argv[])
{
  struct idmap      *pdb;
  struct idmap_elem *pel;
  int i;

  if (argc != 2) {
    err("wrong number of arguments");
    return 1;
  }
  if (!(pdb = read_idmap(argv[1]))) return 1;

  printf("%s", pdb->top_comments);
  for (i = 0; i < pdb->items.u; i++) {
    pel = pdb->items.v[i];

    printf("%s:%d:%s:%s:%s:%s:%s\n",
           pel->login, pel->id, pel->tname,
           pel->login1, pel->room1,
           pel->login2, pel->room2);
  }
  printf("%s", pdb->bottom_comments);

  return 0;
}
#endif

static void
err(char const *format, ...)
{
  va_list args;

  va_start(args, format);
  vfprintf(stderr, format, args);
  va_end(args);
  putc('\n', stderr);
}

/**
 * Local variables:
 *  compile-command: "make"
 *  c-font-lock-extra-types: ("\\sw+_t" "FILE")
 * End:
 */

