/* -*- c -*- */
/* $Id: inetdb.c,v 1.2 2002/03/22 20:35:58 cher Exp $ */

/* Copyright (C) 2001,2002 Alexander Chernov <cher@ispras.ru> */

/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "inetdb.h"

#include <reuse/xalloc.h>

#include <stdio.h>
#include <string.h>
#include <stdarg.h>

static void err(char const *, ...);

enum
{
  LOGIN_FIELD_LEN = 32,
  ID_FIELD_LEN = 32,
  FLAGS_FIELD_LEN = 32,
  TNAME_FIELD_LEN = 64,
  NAME1_FIELD_LEN = 128,
  NAME2_FIELD_LEN = 128,
  NAME3_FIELD_LEN = 128,
  EMAIL_FIELD_LEN = 128,
  COUNTRY_FIELD_LEN = 128,
  ZIP_FIELD_LEN = 128,
  CITY_FIELD_LEN = 128,
  ADDRESS_FIELD_LEN = 128,
  DEST_FIELD_LEN = 128,
  HOMEPAGE_FIELD_LEN = 128
};

struct inetdb_elem
{
  char *login;
  int  id;
  char *flags;
  char *tname;
  char *name1;
  char *name2;
  char *name3;
  char *email;
  char *country;
  char *zip;
  char *city;
  char *address;
  char *dest;
  char *homepage;
};

struct inetdb_elem_arr
{
  int                  a, u;
  struct inetdb_elem **v;
};

struct inetdb
{
  char *top_comments;
  char *bottom_comments;
  struct inetdb_elem_arr items;
  int   id_size;
  int  *id_map;
};

static char b_login[LOGIN_FIELD_LEN];
static char b_id[ID_FIELD_LEN];
static char b_flags[FLAGS_FIELD_LEN];
static char b_tname[TNAME_FIELD_LEN];
static char b_name1[NAME1_FIELD_LEN];
static char b_name2[NAME1_FIELD_LEN];
static char b_name3[NAME1_FIELD_LEN];
static char b_email[EMAIL_FIELD_LEN];
static char b_country[COUNTRY_FIELD_LEN];
static char b_zip[ZIP_FIELD_LEN];
static char b_city[CITY_FIELD_LEN];
static char b_address[ADDRESS_FIELD_LEN];
static char b_dest[DEST_FIELD_LEN];
static char b_homepage[HOMEPAGE_FIELD_LEN];

struct fieldinfo
{
  char *name;
  int   maxlen;
  char *ptr;
};

static struct fieldinfo inetdb_fieldinfo [] =
{
  { "login", LOGIN_FIELD_LEN, b_login },
  { "id", ID_FIELD_LEN, b_id },
  { "flags", FLAGS_FIELD_LEN, b_flags },
  { "tname", TNAME_FIELD_LEN, b_tname },
  { "name1", NAME1_FIELD_LEN, b_name1 },
  { "name2", NAME2_FIELD_LEN, b_name2 },
  { "name3", NAME3_FIELD_LEN, b_name3 },
  { "email", EMAIL_FIELD_LEN, b_email },
  { "country", COUNTRY_FIELD_LEN, b_country },
  { "zip", ZIP_FIELD_LEN, b_zip },
  { "city", CITY_FIELD_LEN, b_city },
  { "address", ADDRESS_FIELD_LEN, b_address },
  { "dest", DEST_FIELD_LEN, b_dest },
  { "homepage", HOMEPAGE_FIELD_LEN, b_homepage },

  { 0, 0, 0 }
};

static int
read_fields(char const *s, int separator, const struct fieldinfo *fi,
            char const *file, int line)
{
  int i, j;
  char const *p = s;

  for (i = 0; fi[i].maxlen && *p; i++) {
    j = 0;
    memset(fi[i].ptr, 0, fi[i].maxlen);
    while (*p != 0 && *p != separator && *p != '\n') {
      if (j >= fi[i].maxlen - 1) {
        err("%s: %d: read_fields: field %s is too long", file,
            line, fi[i].name);
        return -1;
      }
      fi[i].ptr[j++] = *p++;
    }
    if (*p == separator || *p == '\n') p++;
  }
  if (fi[i].maxlen != 0) {
    err("%s: %d: read_fields: too few fields read: %d", file, line, i);
    return -1;
  }
  return i;
}

void
free_inetdb_elem(struct inetdb_elem *pdbe)
{
  if (!pdbe) return;
  xfree(pdbe->login);
  xfree(pdbe->flags);
  xfree(pdbe->tname);
  xfree(pdbe->name1);
  xfree(pdbe->name2);
  xfree(pdbe->name3);
  xfree(pdbe->email);
  xfree(pdbe->country);
  xfree(pdbe->zip);
  xfree(pdbe->city);
  xfree(pdbe->address);
  xfree(pdbe->dest);
  xfree(pdbe->homepage);
  XMEMZERO(pdbe, 1);
  xfree(pdbe);
}

void
free_inetdb(struct inetdb *pdb)
{
  int i;

  xfree(pdb->top_comments);
  xfree(pdb->bottom_comments);
  for (i = 0; i < pdb->items.u; i++) {
    free_inetdb_elem(pdb->items.v[i]);
  }
  if (pdb->items.a > 0) {
    XMEMZERO(pdb->items.v, pdb->items.a);
  }
  xfree(pdb->items.v);
  xfree(pdb->id_map);
  XMEMZERO(pdb, 1);  
}

struct inetdb *
read_inetdb(char const *name)
{
  struct inetdb *pdb = 0;
  struct inetdb_elem *pelem = 0;
  FILE *in = 0;
  int line = 0;
  int len;
  int tid, n, i;
  int state = 0;
  char linebuf[1024];

  if (!(in = fopen(name, "r"))) {
    err("read_localdb: cannot open `%s'", name);
    goto cleanup;
  }
  XCALLOC(pdb, 1);

  while (fgets(linebuf, sizeof(linebuf), in) != NULL) {
    line++;
    len = strlen(linebuf);
    if (len == sizeof(linebuf) - 1 && linebuf[len - 1] != '\n') {
      err("%s: %d: read_localdb: line is too long", name, line);
      goto cleanup;
    }
    if (linebuf[0] == '#') {
      if (state == 0) {
        pdb->top_comments = xstrmerge1(pdb->top_comments, linebuf);
      } else if (state == 2) {
        pdb->bottom_comments = xstrmerge1(pdb->bottom_comments, linebuf);
      }
      if (!strcmp(linebuf, "###\n")) {
        state++;
        if (state == 2)
          pdb->bottom_comments = xstrmerge1(pdb->bottom_comments, linebuf);
      }
      continue;
    }
    if (read_fields(linebuf, ';', inetdb_fieldinfo, name, line) < 0)
      goto cleanup;
    if (sscanf(b_id, "%d %n", &tid, &n) != 1 || b_id[n]
        || tid <= 0 || tid > 65535) {
      err("%s: %d: invalid team_id: %s, %d", name, line, b_id, tid);
      goto cleanup;
    }

    XCALLOC(pelem, 1);
    pelem->login = xstrdup(b_login);
    pelem->id = tid;
    pelem->flags = xstrdup(b_flags);
    pelem->tname = xstrdup(b_tname);
    pelem->name1 = xstrdup(b_name1);
    pelem->name2 = xstrdup(b_name2);
    pelem->name3 = xstrdup(b_name3);
    pelem->email = xstrdup(b_email);
    pelem->country = xstrdup(b_country);
    pelem->zip = xstrdup(b_zip);
    pelem->city = xstrdup(b_city);
    pelem->address = xstrdup(b_address);
    pelem->dest = xstrdup(b_dest);
    pelem->homepage = xstrdup(b_homepage);
    XEXPAND2(pdb->items);
    pdb->items.v[pdb->items.u++] = pelem;
    pelem = 0;
  }

  fclose(in);

  pdb->id_size = -1;
  for (i = 0; i < pdb->items.u; i++) {
    if (pdb->items.v[i]->id > pdb->id_size)
      pdb->id_size = pdb->items.v[i]->id;
  }
  pdb->id_size++;
  if (pdb->id_size > 0) {
    XCALLOC(pdb->id_map, pdb->id_size);
    for (i = 0; i < pdb->id_size; i++)
      pdb->id_map[i] = -1;
  }
  for (i = 0; i < pdb->items.u; i++) {
    pdb->id_map[pdb->items.v[i]->id] = i;
  }

  return pdb;

 cleanup:
  if (in) fclose(in);
  free_inetdb(pdb);
  free_inetdb_elem(pelem);
  return 0;
}

int
inetdb_id_size(struct inetdb *db)
{
  return db->id_size;
}

int
inetdb_get_index(struct inetdb *db, int teamid)
{
  return db->id_map[teamid];
}

char *
inetdb_get_name(struct inetdb *db, int teamid)
{
  return db->items.v[db->id_map[teamid]]->tname;
}

char *
inetdb_get_login(struct inetdb *db, int teamid)
{
  return db->items.v[db->id_map[teamid]]->login;
}

char *
inetdb_get_email(struct inetdb *db, int teamid)
{
  return db->items.v[db->id_map[teamid]]->email;
}

#ifdef ALONE
int
main(int argc, char *argv[])
{
  struct inetdb      *pdb;
  struct inetdb_elem *pel;
  int i;

  if (argc != 2) {
    err("wrong number of arguments");
    return 1;
  }
  if (!(pdb = read_inetdb(argv[1]))) return 1;

  printf("%s", pdb->top_comments);
  for (i = 0; i < pdb->items.u; i++) {
    pel = pdb->items.v[i];

    printf("%s;%d;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s;%s\n",
           pel->login, pel->id, pel->flags, pel->tname,
           pel->name1,
           pel->name2,
           pel->name3,
           pel->email,
           pel->country,
           pel->zip,
           pel->city,
           pel->address,
           pel->dest,
           pel->homepage);
  }
  printf("%s", pdb->bottom_comments);

  return 0;
}
#endif

static void
err(char const *format, ...)
{
  va_list args;

  va_start(args, format);
  vfprintf(stderr, format, args);
  va_end(args);
  putc('\n', stderr);
}

/**
 * Local variables:
 *  compile-command: "make"
 *  c-font-lock-extra-types: ("\\sw+_t" "FILE")
 * End:
 */

